#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert QuantizedLayerNormalization to Akida.
"""
from akida import LayerType, MadNorm
import quantizeml.layers as qlayers
import numpy as np

from .weights import broadcast_and_set_variable


def _set_madnorm_variables(ak_layer, k_layer):
    """Computes and sets the variables for an Akida MadNorm layer.

    This function converts the variables of a Keras layer and sets them into
    the corresponding variables of the equivalent Akida layer.

    Args:
        ak_layer (:obj:`akida.Layer`): the targeted akida layer.
        k_layer (:obj:`tf.keras.Layer`): the source quantized layer.
    """
    assert isinstance(k_layer, qlayers.QuantizedLayerNormalization)
    assert ak_layer.parameters.layer_type == LayerType.MadNorm
    assert isinstance(k_layer.gamma_quantizer, qlayers.WeightQuantizer)
    assert isinstance(k_layer.beta_quantizer, qlayers.AlignedWeightQuantizer)

    # get the QuantizedLayerNormalization gamma and shift
    gamma_ak = k_layer.gamma_quantizer.qweights.value.values.numpy()
    gamma_shift = k_layer.gamma_shift.value.numpy().astype(np.uint8)
    # get the QuantizedLayerNormalization beta and shift
    beta_quantizer = k_layer.beta_quantizer
    beta = beta_quantizer.qweights.value.values.numpy().astype(np.int32)
    beta_shift = beta_quantizer.shift.value.numpy().astype(np.uint8)
    beta_ak = beta >> beta_shift

    variables_ak = ak_layer.variables

    input_shift = k_layer.input_shift.value
    if np.any(input_shift < 0):
        raise RuntimeError(
            f"Layer {k_layer.name} contains negative values for " +
            "input_shift, that is not supported")
    if input_shift is not None:
        broadcast_and_set_variable(variables_ak, "input_shift",
                                   input_shift.numpy().astype(np.uint8))
    variables_ak["gamma"] = gamma_ak.astype(np.int32)
    broadcast_and_set_variable(variables_ak, "gamma_shift", gamma_shift)
    variables_ak["beta"] = beta_ak
    broadcast_and_set_variable(variables_ak, "beta_shift", beta_shift)
    out_quantizer = getattr(k_layer, "out_quantizer", False)
    if out_quantizer:
        assert isinstance(out_quantizer, qlayers.OutputQuantizer)
        qscales = out_quantizer.qscales
        variables_ak["output_scales"] = qscales.value.values.numpy().astype(
            np.uint8)
        variables_ak["output_shift"] = out_quantizer.shift.value.numpy(
        ).astype(np.int8)


def _create_madnorm(layer):
    """Parses a quantizeml QuantizedLayerNormalization layer and returns the
    params to create the corresponding Akida MadNorm layer.

    Args:
        layer (:obj:`tf.keras.Layer`): the quantizeml
            QuantizedLayerNormalization layer to convert.
    """
    assert isinstance(layer, qlayers.QuantizedLayerNormalization)

    # Find out if there is a quantizer
    out_quantizer = getattr(layer, "out_quantizer", False)
    # In quantizeml one bit is reserved automatically for the sign, but in
    # akida this is rather checked during the clipping operations.
    buffer_bits = layer.buffer_bitwidth + 1
    if out_quantizer:
        output_bits = out_quantizer.bitwidth
    else:
        # Default to buffer bitwidth
        output_bits = buffer_bits

    return MadNorm(output_bits=output_bits,
                   buffer_bits=buffer_bits,
                   name=layer.name)


def convert_quantized_madnorm(model_ak, layer_k):
    """Converts QuantizedLayerNormalization layer and its variables and adds
    it to the Akida's model.

    Args:
        model_ak (:obj:`akida.Model`): the Akida model where the model will
            be added.
        layer_k (:obj:`tf.keras.Layer`): the quantizeml
            QuantizedLayerNormalization layer to convert.
    """
    if not isinstance(layer_k, qlayers.QuantizedLayerNormalization):
        raise TypeError(f"Layer {layer_k.name} was expected to be "
                        "QuantizedLayerNormalization")
    layer_ak = _create_madnorm(layer_k)
    model_ak.add(layer_ak)
    _set_madnorm_variables(layer_ak, layer_k)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert QuantizedConcatenate to Akida.
"""
from akida import Concatenate
import quantizeml.layers as qlayers

from .layer_utils import get_inbound_layers_names


def convert_quantized_concatenate(model_ak, layer_k):
    """Converts QuantizedConcatenate layer and its variables and adds it to the
    Akida's model.

    Args:
        layer (:obj:`tf.keras.Layer`): the quantizeml QuantizedConcatenate
            layer to convert.
        :obj:`akida.Model`: the Akida model where the model will be added.
    """
    if not isinstance(layer_k, qlayers.QuantizedConcatenate):
        raise TypeError(f"Layer {layer_k.name} was expected to be "
                        "QuantizedConcatenate")
    # Get inbound layers names
    inbound_layers_names = get_inbound_layers_names(layer_k)
    # Retrieve the akida layers version of the inbound layers.
    inbound_layers_ak = [
        model_ak.get_layer(ly) for ly in inbound_layers_names
    ]
    # Create and add layer to the akida model
    # It is quite simple because there are no variables or params to be set.
    layer_ak = Concatenate()
    model_ak.add(layer_ak, inbound_layers_ak)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert QuantizedDense to Akida.
"""
from akida import LayerType, Dense2D
import quantizeml.layers as qlayers
import numpy as np

from .weights import broadcast_and_set_variable


def _set_dense2d_variables(ak_layer, k_layer):
    """Computes and sets the variables for an Akida Dense2D layer.

    This function converts the variables of a Keras layer and sets them into
    the corresponding variables of the equivalent Akida layer.

    Args:
        layer_ak (:obj:`akida.Layer`): the targeted akida layer.
        k_layer (:obj:`tf.keras.Layer`): the source quantized layer.
    """
    assert isinstance(k_layer, qlayers.QuantizedDense)
    assert ak_layer.parameters.layer_type == LayerType.Dense2D
    assert isinstance(k_layer.weight_quantizer, qlayers.WeightQuantizer)
    assert isinstance(k_layer.bias_quantizer, qlayers.AlignedWeightQuantizer)

    # get the QuantizedDense weights
    weights_ak = k_layer.weight_quantizer.qweights.value.fp.values.numpy()
    # get the QuantizedDense bias and shift
    bias_quantizer = k_layer.bias_quantizer
    bias = bias_quantizer.qweights.value.values.numpy().astype(np.int32)
    bias_shift = bias_quantizer.shift.value.numpy().astype(np.uint8)
    bias_ak = bias >> bias_shift

    variables_ak = ak_layer.variables

    input_shift = k_layer.input_shift.value
    if input_shift is not None:
        broadcast_and_set_variable(variables_ak, "input_shift",
                                   input_shift.numpy().astype(np.uint8))
    variables_ak["weights"] = weights_ak.astype(np.int8)
    variables_ak["bias"] = bias_ak
    variables_ak["bias_shift"] = bias_shift
    out_quantizer = getattr(k_layer, "out_quantizer", False)
    if out_quantizer:
        assert isinstance(out_quantizer, qlayers.OutputQuantizer)
        qscales = out_quantizer.qscales
        variables_ak["output_scales"] = qscales.value.values.numpy().astype(
            np.uint8)
        variables_ak["output_shift"] = out_quantizer.shift.value.numpy(
        ).astype(np.int8)


def _create_dense2d(layer):
    """Parses a quantizeml QuantizedDense layer and returns the params to
    create the corresponding Akida Dense2D layer.

    Args:
        layer (:obj:`tf.keras.Layer`): the quantizeml QuantizedDense layer to
            convert.
        params (dict): will contain the parameters of the future Akida Dense2D
            layer.

    """
    assert isinstance(layer, qlayers.QuantizedDense)
    weight_bits = layer.weight_quantizer.bitwidth
    # The only weight bitwidth supported is 4
    assert weight_bits == 4
    # Find out if there is a quantizer
    out_quantizer = getattr(layer, "out_quantizer", False)
    # In quantizeml one bit is reserved for the sign in the buffer bitwidth
    # variable, but in akida this value has to be added back to have the
    # correct clipping.
    buffer_bits = layer.buffer_bitwidth + 1
    if out_quantizer:
        output_bits = out_quantizer.bitwidth
    else:
        # Default to buffer bitwidth
        output_bits = buffer_bits

    return Dense2D(units=layer.units,
                   output_bits=output_bits,
                   buffer_bits=buffer_bits,
                   name=layer.name)


def convert_quantized_dense(model_ak, layer_k):
    """Converts QuantizedDense layer and its variables and adds it to the
    Akida's model.

    Args:
        layer (:obj:`tf.keras.Layer`): the quantizeml QuantizedDense layer to
            convert.
        :obj:`akida.Model`: the Akida model where the model will be added.
    """
    if not isinstance(layer_k, qlayers.QuantizedDense):
        raise TypeError(f"Layer {layer_k.name} was expected to be "
                        "QuantizedDense")
    layer_ak = _create_dense2d(layer_k)
    model_ak.add(layer_ak)
    _set_dense2d_variables(layer_ak, layer_k)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Utility function to convert layer to akida.
"""


def get_inbound_layers_names(layer):
    """Returns the inbounds layers names of a layer.

    Args:
        layer (:obj:`tf.keras.Layer`): The source layer.

    Returns:
        :list: the inbounds layers names.
    """
    # Get inbound layers names
    inbound_layers_names = []
    in_node = layer._inbound_nodes[0]
    inbound_layers = in_node.inbound_layers
    for layer in inbound_layers:
        inbound_layers_names.append(layer.name)
    return inbound_layers_names



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert Stem block quantized layers to Akida.
Those layers are:
    - The Embedding layer
    - The Reshape layer
    - The ClassToken (+ the DistToken for distilled models) layer(s)
    - The AddPosEmbedding layer
"""
from akida import LayerType, Stem
import quantizeml.layers as qlayers
import numpy as np

from .weights import broadcast_and_set_variables


def stem_layers(model_k):
    """Returns the Stem layers of a quantized transformer model.
    For now the Stem is composed only of the Embedding QuantizedConv2D
    and QuantizedReshape.

    Args:
        model_k (:obj:`keras.Model`): the quantized transformer model.

    Return:
        list: the stem block layers.
    """
    stem_block = []

    # The stem of a transfomer model starts with a QuantizedConv2D layer
    # (i.e layer just after the InputLayer layer), followed by an optionnal
    # QuantizedReshape
    if isinstance(model_k.layers[1], qlayers.QuantizedConv2D):
        layer = model_k.layers[1]
        assert layer.kernel_size == layer.strides
        assert layer.padding == 'valid'
        stem_block.append(layer)
        # add the QuantizedReshape layer to the stem if available
        if len(model_k.layers) > 2 and isinstance(model_k.layers[2],
                                                  qlayers.QuantizedReshape):
            stem_block.append(model_k.layers[2])

    # Extract the QuantizedClassToken and QuantizedAddPositionEmbs layers
    nb_class_tokens = 0
    first_class_token_layer = None
    last_class_token_layer = None
    for idx, layer in enumerate(model_k.layers):
        if isinstance(layer, qlayers.QuantizedClassToken):
            if first_class_token_layer is None:
                first_class_token_layer = idx
            last_class_token_layer = idx
            stem_block.append(layer)
            nb_class_tokens += 1
        elif isinstance(layer, qlayers.QuantizedAddPositionEmbs):
            stem_block.append(layer)

    # the QuantizedClassToken layers should be in a row
    if nb_class_tokens > 0:
        if any(not isinstance(layer, qlayers.QuantizedClassToken) for layer in
               model_k.layers[first_class_token_layer:last_class_token_layer +
                              1]):
            raise TypeError(
                "The QuantizedClassToken layers of the Stem must be in a row.")

    return stem_block


def _set_stem_variables(ak_layer, stem_layers):
    """Computes and sets the variables for an Akida Stem layer.

    This function converts the variables of a Keras layers and sets them into
    the corresponding variables of the equivalent Akida layer.

    Args:
        ak_layer (:obj:`akida.Layer`): the targeted akida layer.
        stem_layers (list): list of the source quantized layers.
    """
    assert ak_layer.parameters.layer_type == LayerType.Stem

    # Note for now only the Embedding layer conversion is handled
    # the Embedding (i.e the QuantizedConv2D layer) should be the
    # first layer of the Stem
    embedding_layer = stem_layers[0]

    cls_layers = []
    add_pos_emb_layer = None
    # Get the QuantizedClassToken layer(s) and the optional
    # QuantizedAddPositionEmbs layer
    for layer in stem_layers:
        if isinstance(layer, qlayers.QuantizedClassToken):
            cls_layers.append(layer)
        elif isinstance(layer, qlayers.QuantizedAddPositionEmbs):
            add_pos_emb_layer = layer

    # Prepare a dict for akida variables
    variables_ak = {}

    # get the Embedding weights
    weights_ak = embedding_layer.weight_quantizer.qweights.value.fp.values.numpy()

    # get the Embedding bias
    bias_quantizer = embedding_layer.bias_quantizer
    bias = bias_quantizer.qweights.value.values.numpy().astype(np.int32)
    bias_shift = bias_quantizer.shift.value.numpy().astype(np.uint8)
    bias_ak = bias >> bias_shift

    if len(cls_layers) > 0:
        tokens_ak = []
        shifts_tokens_ak = []
        for cls_layer in cls_layers:
            # get the ClassToken layer token variable (aka cls member)
            cls_quantizer = cls_layer.cls_quantizer
            token = cls_quantizer.qweights.value.values.numpy().astype(np.int32)
            token_shift = cls_quantizer.shift.value.numpy().astype(np.uint8)
            token_ak = token >> token_shift
            # Insert the token value at the first position. This allows us to match
            # the model concatenation order.
            tokens_ak.insert(0, np.squeeze(token_ak))
            shifts_tokens_ak.insert(0, token_shift)
        variables_ak["tokens"] = np.stack(tokens_ak)
        variables_ak["tokens_shift"] = np.concatenate(shifts_tokens_ak)

    if add_pos_emb_layer:
        # get the positional embedding matrix
        pos_emb_quantizer = add_pos_emb_layer.pe_quantizer
        pos_emb = pos_emb_quantizer.qweights.value.values.numpy().astype(np.int32)
        pos_emb_shift = pos_emb_quantizer.shift.value.numpy().astype(np.uint8)
        pos_emb_ak = pos_emb >> pos_emb_shift
        variables_ak["pos_embedding"] = pos_emb_ak
        # Get the QuantizedAddPositionEmbs layer shifts
        variables_ak["pos_embs_shift"] = pos_emb_shift

    variables_ak["weights"] = weights_ak.astype(np.int8)
    variables_ak["bias"] = bias_ak.astype(np.int32)
    # Get the Embedding layer shifts
    variables_ak["bias_shift"] = bias_shift.astype(np.uint8)

    # The output quantizer corresponds to the output quantizer of the last layer of the stem
    out_quantizer = getattr(stem_layers[-1], "out_quantizer", False)
    if out_quantizer:
        assert isinstance(out_quantizer, qlayers.OutputQuantizer)
        qscales = out_quantizer.qscales
        variables_ak["output_scales"] = qscales.value.values.numpy().astype(
            np.uint8)
        variables_ak["output_shift"] = out_quantizer.shift.value.numpy(
        ).astype(np.int8)

    broadcast_and_set_variables(ak_layer, variables_ak)


def _create_stem(input_shape, layers):
    """Parses the quantizeml quantized layers of the Stem block and returns the
    params to create the corresponding Akida Stem layer.

    Args:
        input_shape (tuple): the input shape of the Stem.
        layers (list): the quantizeml quantized layers of the Stem to convert.
    """
    # Note: for now only the Embedding layer conversion is handled
    embedding_layer = layers[0]

    cls_layers = []
    add_pos_emb_layer = None
    # Get the QuantizedClassToken layer(s) and the optional
    # QuantizedAddPositionEmbs layer
    for layer in layers:
        if isinstance(layer, qlayers.QuantizedClassToken):
            cls_layers.append(layer)
        elif isinstance(layer, qlayers.QuantizedAddPositionEmbs):
            add_pos_emb_layer = layer

    # Find out if there is a quantizer
    out_quantizer = getattr(layers[-1], "out_quantizer", False)
    # In quantizeml one reserves automaticaly one bit for the sign, but in akida
    # this is rather checked during the clipping operations.
    buffer_bits = embedding_layer.buffer_bitwidth + 1
    if out_quantizer:
        output_bits = out_quantizer.bitwidth
    else:
        # Default to buffer bitwidth
        output_bits = buffer_bits

    collapse_spatial_dims = len(layers) > 1 and isinstance(
        layers[1], qlayers.QuantizedReshape)

    num_non_patch_tokens = len(cls_layers)

    # Default tokens_bits at 10
    tokens_bits = 10
    # All the ClassToken layers should normally have the same cls_bitwidth
    if num_non_patch_tokens > 0:
        tokens_bits = cls_layers[0].cls_quantizer.bitwidth

    # Handle optional AddPositionEmbedding params
    add_pos_embs_available = False
    pos_embedding_bits = 8
    if add_pos_emb_layer:
        add_pos_embs_available = True
        pos_embedding_bits = add_pos_emb_layer.pe_quantizer.bitwidth

    return Stem(input_shape=input_shape,
                filters=embedding_layer.filters,
                kernel_size=embedding_layer.kernel_size[0],
                weights_bits=embedding_layer.weight_quantizer.bitwidth,
                bias_bits=embedding_layer.bias_quantizer.bitwidth,
                tokens_bits=tokens_bits,
                pos_embedding_bits=pos_embedding_bits,
                output_bits=output_bits,
                buffer_bits=buffer_bits,
                collapse_spatial_dims=collapse_spatial_dims,
                num_non_patch_tokens=num_non_patch_tokens,
                add_pos_embs_available=add_pos_embs_available,
                name="Stem")


def convert_quantized_stem_layers(model_ak, input_shape, layers_k):
    """Converts QuantizedDense layer and its variables and adds it to the
    Akida's model.

    Args:
        model_ak (:obj:`akida.Model`): the Akida model where the model will be added.
        input_shape (tuple): the input shape of the Stem.
        layers_k (list): the quantizeml quantized layers of the Stem to convert.
    """
    # Note: for now only the Embedding layer conversion is handled
    embedding_layer = layers_k[0]
    if not isinstance(embedding_layer, qlayers.QuantizedConv2D):
        raise TypeError(f"Layer {embedding_layer.name} was expected to be "
                        "QuantizedConv2D")

    layer_ak = _create_stem(input_shape, layers_k)
    model_ak.add(layer_ak)
    _set_stem_variables(layer_ak, layers_k)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert QuantizedAdd to Akida.
"""
from akida import LayerType, Add
import quantizeml.layers as qlayers
import numpy as np

from .layer_utils import get_inbound_layers_names
from .weights import broadcast_and_set_variable


def _set_add_variables(ak_layer, k_layer):
    """Computes and sets the variables for an Akida Add layer.

    This function converts the variables of a Keras layer and sets them into
    the corresponding variables of the equivalent Akida layer.

    Args:
        layer_ak (:obj:`akida.Layer`): the targeted akida layer.
        k_layer (:obj:`tf.keras.Layer`): the source quantized layer.
    """
    assert isinstance(k_layer, qlayers.QuantizedAdd)
    assert ak_layer.parameters.layer_type == LayerType.Add

    variables_ak = ak_layer.variables

    a_shift = k_layer.a_shift.value.numpy().astype(np.uint8)
    broadcast_and_set_variable(variables_ak, "a_shift", a_shift)
    b_shift = k_layer.b_shift.value.numpy().astype(np.uint8)
    broadcast_and_set_variable(variables_ak, "b_shift", b_shift)

    out_quantizer = getattr(k_layer, "out_quantizer", False)
    if out_quantizer:
        assert isinstance(out_quantizer, qlayers.OutputQuantizer)
        qscales = out_quantizer.qscales
        # If inputs were FixedPoint, there will be no scales
        if qscales.value:
            out_scales = qscales.value.values.numpy().astype(np.uint8)
            variables_ak["output_scales"] = out_scales
        out_shift = out_quantizer.shift.value.numpy().astype(np.int8)
        variables_ak["output_shift"] = out_shift


def _create_add(layer):
    """Parses a quantizeml QuantizedAdd layer and returns the corresponding
    Akida Add layer.

    Args:
        layer (:obj:`tf.keras.Layer`): the quantizeml QuantizedAdd layer to
            convert.
    """
    assert isinstance(layer, qlayers.QuantizedAdd)
    # In quantizeml one reserves automaticaly one bit for the sign, but in akida
    # this is rather checked during the clipping operations.
    buffer_bits = layer.buffer_bitwidth + 1
    if getattr(layer, 'out_quantizer', False):
        output_bits = layer.out_quantizer.bitwidth
    else:
        output_bits = buffer_bits
    return Add(output_bits=output_bits,
               buffer_bits=buffer_bits,
               name=layer.name)


def convert_quantized_add(model_ak, layer_k):
    """Converts QuantizedAdd layer and its variables and adds it to the
    Akida's model.

    Args:
        layer (:obj:`tf.keras.Layer`): the quantizeml QuantizedAdd layer to
            convert.
        :obj:`akida.Model`: the Akida model where the model will be added.
    """
    if not isinstance(layer_k, qlayers.QuantizedAdd):
        raise TypeError(f"Layer {layer_k.name} was expected to be "
                        "QuantizedAdd")
    # Get inbound layers names
    inbound_layers_names = get_inbound_layers_names(layer_k)
    # Retrieve the akida layers version of the inbound layers.
    inbound_layers_ak = [
        model_ak.get_layer(ly) for ly in inbound_layers_names
    ]
    # Create and add layer to the akida model
    layer_ak = _create_add(layer_k)
    model_ak.add(layer_ak, inbound_layers_ak)
    # Set the akida layer converted variables
    _set_add_variables(layer_ak, layer_k)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Parsing functions to generate an Akida model from a Keras model quantized with quantizeml api.
"""

import quantizeml.layers as qlayers
from quantizeml.models import record_quantization_variables
from akida import Model

from .dense2d import convert_quantized_dense
from .add import convert_quantized_add
from .input_data import convert_input
from .shiftmax import convert_quantized_shiftmax
from .attention import convert_quantized_attention
from .madnorm import convert_quantized_madnorm
from .stem import convert_quantized_stem_layers, stem_layers
from .concatenate import convert_quantized_concatenate


def generate_model(model, input_is_image):
    """Generates an Akida model.

    This function creates an Akida model by iterating through the layers of the
    quantized model. For each layer, the corresponding akida layer is created and
    added to the Akida model.

    Args:
        model (:obj:`tf.keras.Model`): a Keras model to convert.
        input_is_image (bool): True if input is an 8-bit unsigned tensors (like images).

    Returns:
        :obj:`akida.Model`: the generated Akida model.

    """

    # First store necessary variables for conversion
    record_quantization_variables(model)

    model_ak = Model()

    # Convert the stem block layer if the keras model has one, otherwise convert the
    # keras InputLayer into an InputData layer
    stem_block = stem_layers(model)
    if len(stem_block) != 0:
        input_shape = model.input_shape[1:]
        convert_quantized_stem_layers(model_ak, input_shape, stem_block)
    else:
        convert_input(model_ak, model.layers[0], input_is_image)

    # Convert the remaining layers
    for layer in model.layers[len(stem_block) + 1:]:
        # Create and add layer to the akida model
        if isinstance(layer, qlayers.QuantizedDense):
            convert_quantized_dense(model_ak, layer)
        elif isinstance(layer, qlayers.QuantizedAdd):
            convert_quantized_add(model_ak, layer)
        elif isinstance(layer, qlayers.QuantizedShiftmax):
            convert_quantized_shiftmax(model_ak, layer)
        elif isinstance(layer, qlayers.QuantizedAttention):
            convert_quantized_attention(model_ak, layer)
        elif isinstance(layer, qlayers.QuantizedLayerNormalization):
            convert_quantized_madnorm(model_ak, layer)
        elif isinstance(layer, qlayers.QuantizedConcatenate):
            convert_quantized_concatenate(model_ak, layer)
        elif isinstance(layer, qlayers.QuantizedRescaling):
            # Ignore QuantizedRescaling layer, not necessary in Akida.
            continue
        else:
            # If you got here, the layer is not supported: raise an error.
            raise RuntimeError(f"Layer {layer.name}: unsupported type "
                               f"{layer.__class__.__name__}.")

    return model_ak
